{
        Program Name : MailSlot
        Package Name : LB
        Purpose      : Windows mailslot
        Institution  : Conroy & Conroy Co.
        Date Written : 11-June-2002
        Written By   : Alan Conroy
        Version      : 1.0

	Copyright (C) 2002 by Alan Conroy.  Released to the public domain.

        TO THE GLORY OF GOD THROUGH JESUS CHRIST

        *************************************************************
        *                                                           *
        *        M O D I F I C A T I O N   H I S T O R Y            *
        *                                                           *
        *    DATE        BY          REASON                         *
        *                                                           *
        *                                                           *
        *************************************************************

        *************************************************************
        *                                                           *
        *        P R O G R A M   P U R P O S E                      *
        *                                                           *
        *************************************************************

        Introduction
        ------------
          TMail_Slot encapsulates a windows mail slot.  It is designed for
        client/server communications, but can be used peer-to-peer also.

          In the following descriptions, "Mail Slot" refers to an instance of a
        TMail_Slot.  "mailslot" refers to an MS Windows mailslot.

        General operation
        -----------------
        A single server mail slot is created with a specific name (let's call it
        "S").  Any number of client mail slots are created with the name of the
        server slot, thus indicating who their server is.  As each client is
        created, they request a unique ID from the server.  This ID is then used
        to create a Windows mailslot unique to that client.  The server has a
        single Windows mailslot for all incoming messages.  The clients all
        write to the server's Windows mailslot, while reading from their own
        unique Windows mailslot.  Given two clients (C1, and C2), this is what
        the configuration looks like:

                Mail slot Name      Incoming         Outgoing
                --------------      --------         --------
                S                   mailslot\S       varies
                C1                  mailslot\S_0     mailslot\S
                C2                  mailslot\S_1     mailslot\S

        When the server sends a message to a particular client, it uses the
        appropriate client's incoming mailslot as it's outgoing mailslot.  Thus,
        each mailslot is a single-direction data "pipe".


        Connection protocol
        -------------------
        For a client to get a response back from the server when connecting, so
        it can ask for a unique ID, it must have a Windows mailslot it can use
        to read data from the server.  All clients connecting to a server use
        the mailslot\S_Client Windows mailslot.  To ensure that only one client
        at a time has this open, the first step is to set an exclusive mutex.
        Then an ONLINE command is sent to the server, who responds to the
        mailslot\S_Client with the unique ID (via an ASSIGN command).  The
        client then closes mailslot\S_Client and releases the mutex, opening its
        now unique Windows mailslot.  Summary of steps:
                1. Wait for availability of mutex names "S_Client".
                2. Create mutex.
                3. Create mailslot\S_Client
                4. Send ONLINE command to server
                5. Wait for server response.
                6. Get ASSIGN command from mailslot\S_Client.
                7. Open unique mailslot.
                8. close mailslot\S_Client
                9. Release mutex.

        Disconnection protocol is simpler since the client has a unique mailslot
        at that point.  The client simply sends an "OFFLINE" command to the
        server.


        Normal Message protocol
        -----------------------
        The ONLINE and ASSIGN messages have unique handling because of the
        connection protocol.  Under ALL other circumstances, however, messages
        always consist of a 256-byte header which identifies the source of the
        message, and the message type.  All data following these 256 bytes is
        the actual message data.  The format of the message data is message
        dependant.  There are two types of message: command and text.  They
        differ only in that the header indicates a different type.
}

unit MailSlot ;

interface

uses // Borland...
     Windows, // THandle
     SysUtils, // strpas
     Classes, // TThread

     // Other...
     UEHDefs ; // TUEC

const // TMail_Slot component error codes
      MailSlotErr_Facility = 56 ;
      MailSlotErr_Success = 0 ; // Success
      MailSlotErr_Cannot_Start_Timer = 1 ; // Cannot start Timer Thread
      MailSlotErr_Cannot_Open_Reply = 2 ; // Cannot open reply Mail Slot
      MailSlotErr_Cannot_Open_Remote = 3 ; // Cannot open Remote Mail Slot
      MailSlotErr_Already_Active = 4 ; // Mail slot already active
      MailSlotErr_Not_Active = 5 ; // Mail slot not active

type TMail_Slot = class ;
     TSignal_Thread = class( TThread )
                          private // Instance data...
                              _Mail_Slot : TMail_Slot ;

                          public // Constructors and destructors...
                              constructor Create( Mail_Slot : TMail_Slot ) ;

                          protected  // Overrides...
                              procedure Execute ; override ;
                      end ;
     TMessage_Type = ( MT_Data, MT_Command ) ;
     TMessage_Header = record
                           case boolean of
                               True: ( Fill : byte ;
                                       Computer : string[ 128 ] ; // Name of source computer
                                       Name : string[ 64 ] ; // Name of reply mail slot
                                       PID : integer ; // ID of source process
                                       Typ : TMessage_Type ;
                                       Filler : array[ 0..63 ] of byte ; // More than needed to fill up 256 bytes
                                      ) ;
                               False : ( Msg : string[ 255 ] ) ;
                       end ;

  TMail_Slot_Message_Event = procedure ( Sender : TObject ;
      const Origin, Name, Data : string ) of object ;
  TMail_Slot_Command_Event = procedure ( Sender : TObject ;
      const Origin, Name : string ; var Data : string ) of object ;

  TMail_Slot = class
                   private // Instance data...
                       Header : TMessage_Header ;
                       In_Read_Message : boolean ; // True while processing a message
                       _Interval : word ; // Timer interval
                       Message_Count : DWORD ; // Number of messages waiting in mail slot
                       Next_Client : integer ; // One-up number for assigning unique mail slots
                       Next_Size : DWORD ; // Size of next message in mail slot
                       Read_Handle : THandle ; // Handle from which we read incoming messages
                       Server_Slot_Name : string ; // Name of mail slot for server
                       Signal_Thread : TSignal_Thread ;
                       Slot_Name : string ; // Name of mail slot in which we receive messages
                       Write_Handle : THandle ; // Server's handle
                       Timeout : integer ;

                   public // Constructors and destructors...
                       { Construct mail slot for server named Name.  If we are
                         the server, Server is True.  Otherwise we are a client
                         that communicates with the named server. }
                       constructor Create( const Name : string ;
                           Server : boolean ) ;
                       destructor Destroy ; override ;

                   private // Internal utility routines...
                       function _Activate : boolean ;
                       procedure _Deactivate ;
                       procedure Read_Message ;

                       { Send command Command to receipient Recipient. }
                       procedure Send_Command( const Recipient, Command : string ) ;

                       { Physical transfer of data to recipient }
                       procedure Send_Data( const Recipient : string ;
                           Data : string ) ;

                   protected // Property handlers...
                       procedure Set_Interval( Value : word ) ;
                       procedure Set_Name( const Value : string ) ;

                   protected // Hooks for descendants...
                       procedure Do_Error( Facility, Code : integer ) ; virtual ;
                       procedure Do_Command( const Sender, Name : string ;
                           var Data : string ) ; virtual ;
                       procedure Do_Message( const Sender, Name, Data : string ) ;
                           virtual ;

                       procedure Process_Command( const MsgSender, Name : string ;
                           S : string ) ; virtual ;

                   public // API...
                       Last_Error : TUEC ;
                       On_Command : TMail_Slot_Command_Event ;
                       On_Message : TMail_Slot_Message_Event ;
                       On_Error : TNotifyEvent ;

                       { Activates the mail slot.  No messages can be sent until
                         the mail slot has been activated. }
                       procedure Activate ;

                       { Deactivates the mail slot.  No messages can be sent
                         until the mail slot is avtivated again. }
                       procedure DeActivate ;

                       { Send data in Text to recipient Recipient }
                       procedure Send( const Recipient, Text : string ) ;

                       { Notifies mail slot that called procedure is waiting for
                         a message.  This will allow another message to be
                         processed during the processing of a previous message.
                         The mail slot does not actually wait - the procedure
                         returns immediately. }
                       procedure Wait_For_Message ;

                       { How long to sleep between message checks (ms) }
                       property Interval : word
                           read _Interval
                           write Set_Interval ;

                       property Name : string
                           read Slot_Name
                           write Set_Name ;
               end ; // TMail_Slot

implementation

uses // Borland...
     Forms, // Application

     // Other...
     ASCIIDef, // CRLF
     Standard ; // PID

// TSignal_Thread methods...

// Constructors and destructors...

constructor TSignal_Thread.Create( Mail_Slot : TMail_Slot ) ;

Begin
    Inherited Create( False ) ;

    Priority := tpNormal ;
    _Mail_Slot := Mail_Slot ;
end ;


// Overrides...

procedure TSignal_Thread.Execute ;

begin
    while( not Terminated ) do
    begin
        try
            GetMailSlotInfo( _Mail_Slot.Read_Handle, nil, _Mail_Slot.Next_Size,
                @_Mail_Slot.Message_Count, nil ) ;
            if(
                ( _Mail_Slot.Next_Size <> MAILSLOT_NO_MESSAGE )
                and
                ( _Mail_Slot.Next_Size <> 0 )
              ) then
            begin
                _Mail_Slot.Read_Message ;
            end else
            begin
                Sleep( _Mail_Slot._Interval ) ;
                Application.ProcessMessages ;
            end ;
        except
        end ;
    end ;
end ;


// TMail_Slot methods...

// Constructors and destructors...

constructor TMail_Slot.Create( const Name : string ; Server : boolean ) ;

var Temp : array[ 0..255 ] of char ;
    Len : DWORD ;

begin
    Slot_Name := Name ;
    Server_Slot_Name := Name ;
    _Interval := 0 ;
    Signal_Thread := nil ;
    Len := 255 ;
    GetComputerName( temp, Len ) ;
    Header.Computer := StrPas( temp ) ;
    Len := 255 ;
    if( Server ) then
    begin
        Header.PID := 0 ;
    end else
    begin
        Header.PID := Standard.PID ;
    end ;
    Header.Fill := 255 ;
    Timeout := 5000 ;
end ;


destructor TMail_Slot.Destroy ;

begin
    if( ( Read_Handle <> 0 ) or ( Write_Handle <> 0 ) or ( Signal_Thread <> nil ) ) then
    begin
        DeActivate ;
    end ;

    inherited Destroy ;
end ;


// Internal utility routines...

function TMail_Slot._Activate : boolean ;

begin
    Result := False ; // Assume failure
    Signal_Thread := TSignal_Thread.Create( Self ) ;
    if( Signal_Thread = nil ) then
    begin
        Do_Error( MailSlotErr_Facility, MailSlotErr_Cannot_Start_Timer ) ;
        exit ;
    end ;
    if( Header.PID = 0 ) then // Server
    begin
        Read_Handle := CreateMailSlot( PChar( '\\.\mailslot\' + Slot_Name ), 0, 0, nil ) ;
        Write_Handle := Read_Handle ;
        if( Read_Handle = INVALID_HANDLE_VALUE ) then
        begin
            Signal_Thread.Terminate ;
            Signal_Thread := nil ;
            Do_Error( 0, GetLastError ) ;
            exit ;
        end ;
    end else
    begin // Client
        Read_Handle := 0 ;
        Write_Handle := CreateFile( PChar( '\\.\mailslot\' + Server_Slot_Name ),
            GENERIC_WRITE, FILE_SHARE_READ or FILE_SHARE_WRITE, nil, OPEN_EXISTING,
            FILE_ATTRIBUTE_NORMAL, 0 ) ;
        if( Write_Handle = INVALID_HANDLE_VALUE ) then
        begin
            Signal_Thread.Terminate ;
            Signal_Thread := nil ;
            Do_Error( 0, GetLastError ) ;
            exit ;
        end ;
        if( Server_Slot_Name <> Slot_Name ) then
        begin
            Read_Handle := CreateMailSlot( PChar( '\\.\mailslot\' + Slot_Name ), 0, 0, nil ) ;
        end ;
    end ;
    Result := True ; // Success
end ; // TMail_Slot._Activate


procedure TMail_Slot._Deactivate ;

begin
    if( Signal_Thread <> nil ) then
    begin
        Signal_Thread.Terminate ;
        Signal_Thread := nil ;
    end ;
    if( Read_Handle <> Write_Handle ) then
    begin
        CloseHandle( Read_Handle ) ;
    end ;
    Read_Handle := 0 ;
    CloseHandle( Write_Handle ) ;
    Write_Handle := 0 ;
    Slot_Name := Server_Slot_Name ;
    Header.Name := '' ;
end ;


procedure TMail_Slot.Process_Command( const MsgSender, Name : string ;
    S : string ) ;

var Handle : THandle ;
    Len : DWORD ;

begin
    Do_Command( MsgSender, Name, S ) ;
    if( Header.PID = 0 ) then // Server
    begin
        if( S = 'ONLINE' ) then
        begin
            // Client sent us an ONLINE command.  Respond with a slot assignment
            S := 'ASSIGN ' + Slot_Name + '_' + inttostr( Next_Client ) ;
            Handle := CreateFile( PChar( '\\.\mailslot\' + Slot_Name + '_REPLY' ),
                GENERIC_WRITE, FILE_SHARE_READ or FILE_SHARE_WRITE, nil, OPEN_EXISTING,
                FILE_ATTRIBUTE_NORMAL, 0 ) ;
            if( Handle = INVALID_HANDLE_VALUE ) then
            begin
                Do_Error( MailSlotErr_Facility, MailSlotErr_Cannot_Open_Reply ) ;
                exit ;
            end ;
            WriteFile( Handle, Pointer( S )^, Length( S ), Len, nil ) ;
            inc( Next_Client ) ;
            closehandle( Handle ) ;
        end else
        if( S = 'OFFLINE' ) then
        begin
            // Client is disconnecting.  We do nothing, so ignore the message
        end ;
    end else
    begin
        if( S = 'ASSIGN' ) then
        begin
            // Ignore this, since it should only come in response to our out-going ONLINE command
        end ;
    end ;
end ;


procedure TMail_Slot.Read_Message ;

var Header : TMessage_Header ;
    NewLine : string ;
    Size : DWORD ;

begin
    if( In_Read_Message ) then // Process only one message at a time
    begin
        exit ;
    end ;
    In_Read_Message := True ;
    try
        SetLength( NewLine, Next_Size ) ;
        ReadFile( Read_Handle, PChar( NewLine )^, Next_Size, Size, nil ) ;
        if( length( NewLine ) < 256 ) then // Invalid message
        begin
            exit ;
        end ;
        move( NewLine[ 1 ], Header.Msg, 256 ) ;
        if( length( Header.Msg ) <> 255 ) then
        begin
            exit ; // Invalid message
        end ;
        NewLine := copy( NewLine, 257, length( NewLine ) ) ;
        case Header.Typ of
            MT_Data : Do_Message( Header.Computer, Header.Name, NewLine ) ;
            MT_Command : Process_Command( Header.Computer, Header.Name, NewLine ) ;
        end ;
    finally
        In_Read_Message := False ;
    end ;
end ;


procedure TMail_Slot.Send_Data( const Recipient : string ; Data : string ) ;

var Dummy : integer ;
    Len : DWORD ;
    Handle : THandle ;
    Remote_Computer : string ;

begin
    Data := Header.Msg[ 0 ] + Header.Msg + Data ;
    if( Header.PID = 0 ) then // Server
    begin
        Dummy := pos( '\', Recipient ) ;
        Remote_Computer := copy( Recipient, 1, Dummy - 1 ) ;

        // The server writes to many different handles, so get a handle to this destination
        if( Remote_Computer = Header.Computer ) then
        begin
            Handle := CreateFile( PChar( '\\.\mailslot\' + copy( Recipient, Dummy + 1, length( Recipient ) ) ),
                GENERIC_WRITE, FILE_SHARE_READ or FILE_SHARE_WRITE, nil, OPEN_EXISTING,
                FILE_ATTRIBUTE_NORMAL, 0 ) ;
        end else
        begin
            Handle := CreateFile( PChar( '\\' + Remote_Computer +
                '\mailslot\' + copy( Recipient, Dummy + 1, length( Recipient ) ) ),
                GENERIC_WRITE, FILE_SHARE_READ or FILE_SHARE_WRITE, nil, OPEN_EXISTING,
                FILE_ATTRIBUTE_NORMAL, 0 ) ;
        end ;
        if( Handle = INVALID_HANDLE_VALUE ) then
        begin
            Do_Error( MailSlotErr_Facility, MailSlotErr_Cannot_Open_Remote ) ;
            exit ;
        end ;
        WriteFile( Handle, Pointer( Data )^, Length( Data ), Len, nil ) ;
        closehandle( Handle ) ;
    end else
    begin
        WriteFile( Write_Handle, Pointer( Data )^, Length( Data ), Len, nil ) ;
    end ;
end ;


procedure TMail_Slot.Send_Command( const Recipient, Command : string ) ;

begin
    Header.Typ := MT_Command ;
    Send_Data( Recipient, Command ) ;
end ;


// Property handlers...

procedure TMail_Slot.Set_Interval( Value : word ) ;

begin
   if( _Interval <> Value ) then
   begin
       _Interval := Value ;
   end ;
end ;


procedure TMail_Slot.Set_Name( const Value : string ) ;

begin
    if( Slot_Name <> Value ) then
    begin
        Slot_Name := Value ;
        if( Write_Handle <> 0 ) then
        begin
            DeActivate ;
            Activate ;
        end ;
    end ;
end ;


// Hooks for descendants...

procedure TMail_Slot.Do_Error( Facility, Code : integer ) ;

begin
    Last_Error.Facility := Facility ;
    Last_Error.Code := Code ;
    if Assigned( On_Error ) then
    begin
        On_Error( Self ) ;
    end ;
end ;


procedure TMail_Slot.Do_Command( const Sender, Name : string ; var Data : string ) ;

begin
    if Assigned( On_Command ) then
    begin
        On_Command( Self, Sender, Name, Data ) ;
    end ;
end ;


procedure TMail_Slot.Do_Message( const Sender, Name, Data : string ) ;

begin
    if Assigned( On_Message ) then
    begin
        On_Message( Self, Sender, Name, Data ) ;
    end ;
end ;


// API...

procedure TMail_Slot.Activate ;

var Count : integer ;
    Handle, Mutex : THandle ;
    Actual_Size, Size : DWORD ;
    Start_Time : integer ;
    Temp : string ;

begin
    if( Write_Handle <> 0 ) then
    begin
        Do_Error( MailSlotErr_Facility, MailSlotErr_Already_Active ) ;
        exit ;
    end ;
    if( not _Activate ) then
    begin
        exit ;
    end ;

    if( Header.PID <> 0 ) then // We are a client
    begin
        // Notify the server...

        // First, create exclusive mutex
        Mutex := CreateMutex( nil, False, PChar( Server_Slot_Name + '_CLIENT' ) ) ;
        while( GetLastError = ERROR_ALREADY_EXISTS ) do
        begin
            if( Mutex <> ERROR_INVALID_HANDLE ) then
            begin
                CloseHandle( Mutex ) ;
            end ;
            Sleep( _Interval ) ;
            Mutex := CreateMutex( nil, True, PChar( Server_Slot_Name + '_CLIENT' ) ) ;
        end ;

        // Second, create a server-specific reply mailslot...
        Handle := CreateMailSlot( PChar( '\\.\mailslot\' + Server_Slot_Name + '_REPLY' ), 0, 0, nil ) ;
        if( Handle = INVALID_HANDLE_VALUE ) then
        begin
            Handle := 0 ;
        end ;

        // Third, notify server that we are online
        Send_Command( '.', 'ONLINE' ) ;

        // Fourth, get reply from server that tells us which mailslot to use...
        Start_Time := GetTickCount ;
	if( Handle <> 0 ) then
        begin
            Count := 0 ;
            while( Count = 0 ) do
            begin
                Sleep( _Interval ) ;
                Application.ProcessMessages ;
                GetMailSlotInfo( Handle, nil, Size, @Count, nil ) ;
                if( Size <> MAILSLOT_NO_MESSAGE ) then
                begin
                    break ;
                end ;
                if( GetTickCount > Start_Time + Timeout ) then
                begin
                    // We timed-out, try again
                    Send_Command( '.', 'ONLINE' ) ;
                    Start_Time := GetTickCount ;
                end ;
            end ;
            SetLength( Temp, Size ) ;
            ReadFile( Handle, PChar( Temp )^, Size, Actual_Size, nil ) ;
            if( ( Actual_Size = Size ) and ( copy( Temp, 1, 7 ) = 'ASSIGN ' ) ) then
            begin
                _Deactivate ;
                Slot_Name := copy( Temp, 8, length( Temp ) ) ; // Our new mail slot name
                Header.Name := Slot_Name ;
                _Activate ;
            end ;

            // Release server mailslot
            CloseHandle( Handle ) ;
        end ;

        // Release mutex
        CloseHandle( Mutex ) ;
    end ;
end ; // TMail_Slot.Activate


procedure TMail_Slot.DeActivate ;

begin
    if( Write_Handle = 0 ) then
    begin
        Do_Error( MailSlotErr_Facility, MailSlotErr_Not_Active ) ;
        exit ;
    end ;
    if( Header.PID <> 0 ) then // We are a client
    begin
        Send_Command( '.', 'OFFLINE' ) ;
    end ;
    _Deactivate ;
end ;


procedure TMail_Slot.Send( const Recipient, Text : string ) ;

begin
    Header.Typ := MT_Data ;
    Send_Data( Recipient, Text ) ;
end ;


procedure TMail_Slot.Wait_For_Message ;

begin
    In_Read_Message := False ; // Allow more message processing
end ;


end.
